@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1996-2021 The Squid Software Foundation and contributors
+ * Copyright (C) 1996-2024 The Squid Software Foundation and contributors
  *
  * Squid software is distributed under GPLv2+ license and includes
  * contributions from numerous individuals and organizations.
@@ -10,8 +10,13 @@
  * AUTHOR: Flavio Pescuma. <flavio@marasystems.com>
  */
 
+#ifndef SQUID_SRC_AUTH_DIGEST_LDAP_LDAP_BACKEND_H
+#define SQUID_SRC_AUTH_DIGEST_LDAP_LDAP_BACKEND_H
+
 #include "auth/digest/LDAP/digest_common.h"
 
 extern int LDAPArguments(int argc, char **argv);
 extern void LDAPHHA1(RequestData * requestData);
 
+#endif /* SQUID_SRC_AUTH_DIGEST_LDAP_LDAP_BACKEND_H */
+
